# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing the CI"""

import os
import tempfile

import numpy as np
import pytest
from lxml import etree

from ci.script_utils import files_path, rasters_path, s3_env, vectors_path
from sertit import ci, path, rasters, rasters_rio, vectors

ci.reduce_verbosity()


@s3_env
def test_assert_base():
    """Test CI functions"""
    # assert_val
    ci.assert_val("a", "a", "same string")
    ci.assert_val(None, None, "both None")
    ci.assert_val(np.nan, np.nan, "Nans")
    ci.assert_val([np.nan, np.nan], [np.nan, np.nan], "different vals (list of nans)")
    with pytest.raises(AssertionError):
        ci.assert_val("a", "b", "different string")

    with pytest.raises(AssertionError):
        ci.assert_val("a", None, "only one None")

    # assert_field
    dict_1 = {
        "a": 1,
    }
    dict_2 = {
        "a": 2,
    }
    ci.assert_field(dict_1, dict_1, "a")
    with pytest.raises(AssertionError):
        ci.assert_field(dict_1, dict_2, "a")

    ci.assert_val([1, 2], [1, 2], "same list")
    with pytest.raises(AssertionError):
        ci.assert_val([1, 2], [3, 2], "different list")

    with pytest.raises(AssertionError):
        ci.assert_val(None, 3, "different vals (None)")

    with pytest.raises(AssertionError):
        ci.assert_val(np.nan, 3, "different vals (nans)")

    with pytest.raises(AssertionError):
        ci.assert_val(np.nan, None, "different vals (nans + None)")


@s3_env
def test_assert_dir():
    """Test CI functions"""
    # Dirs
    dir2 = files_path().joinpath("core")

    ci.assert_dir_equal(files_path(), files_path())
    with pytest.raises(AssertionError):
        ci.assert_dir_equal(files_path(), dir2)


@s3_env
def test_assert_files():
    """Test CI functions"""
    ok_path = files_path().joinpath("productPreview.html")
    false_path = files_path().joinpath("false.html")

    ci.assert_files_equal(ok_path, ok_path)
    ci.assert_files_equal(str(ok_path), str(ok_path))
    with pytest.raises(AssertionError):
        ci.assert_files_equal(ok_path, false_path)


@s3_env
def test_assert_vect():
    """Test CI functions"""
    # Vector
    vector_path = vectors_path().joinpath("aoi.geojson")
    vector2_path = vectors_path().joinpath("aoi2.geojson")

    vec_df = vectors.read(vector_path)
    vec2_df = vectors.read(vector2_path)

    assert not vec_df.empty
    ci.assert_geom_equal(vec_df, vector_path)
    with pytest.raises(AssertionError):
        ci.assert_geom_equal(vector_path, vec2_df)

    assert not vec_df.empty
    ci.assert_geom_almost_equal(vec_df, vector_path)
    with pytest.raises(AssertionError):
        ci.assert_geom_almost_equal(vector_path, vec2_df)


@s3_env
def test_assert_raster():
    # Rasters
    raster_path = rasters_path().joinpath("raster.tif")
    raster2_path = rasters_path().joinpath("raster_masked.tif")

    ci.assert_raster_equal(raster_path, raster_path)
    with pytest.raises(AssertionError):
        ci.assert_raster_equal(raster_path, raster2_path)

    # Rasters almost equal
    with tempfile.TemporaryDirectory() as tmp:
        # Read and convert raster to float
        arr, meta = rasters_rio.read(raster_path)
        arr = arr.astype(float)
        meta["dtype"] = float
        raster_float_path = os.path.join(tmp, "raster_float.tif")
        rasters_rio.write(arr, meta, raster_float_path)

        # Slightly change it
        offset = 1e-07
        arr += offset
        meta["transform"].shear(offset, offset)
        raster_almost_path = os.path.join(tmp, "raster_almost.tif")
        rasters_rio.write(arr, meta, raster_almost_path)

        # Almost equal
        ci.assert_raster_almost_equal(raster_float_path, raster_almost_path)
        with pytest.raises(AssertionError):
            raster2_path = rasters_path().joinpath("raster_masked.tif")
            ci.assert_raster_almost_equal(raster_path, raster2_path)

        with pytest.raises(AssertionError):
            offset = 1e-05
            arr += offset
            meta["transform"].shear(offset, offset)
            raster_too_much_path = os.path.join(tmp, "raster_too_much.tif")
            rasters_rio.write(arr, meta, raster_too_much_path)
            ci.assert_raster_almost_equal(raster_float_path, raster_too_much_path)

        # Max mismatch
        ci.assert_raster_max_mismatch(raster_path, raster_path, max_mismatch_pct=0.001)
        with pytest.raises(AssertionError):
            ci.assert_raster_max_mismatch(raster_float_path, raster_almost_path)

        # Magnitude
        ci.assert_raster_almost_equal_magnitude(raster_float_path, raster_almost_path)
        with pytest.raises(AssertionError):
            raster2_path = rasters_path().joinpath("raster_masked.tif")
            ci.assert_raster_almost_equal_magnitude(raster_path, raster2_path)

        with pytest.raises(AssertionError):
            offset = 1e-02
            arr += offset
            meta["transform"].shear(offset, offset)
            raster_too_much_path = os.path.join(tmp, "raster_too_much.tif")
            rasters_rio.write(arr, meta, raster_too_much_path)
            ci.assert_raster_almost_equal_magnitude(
                raster_float_path, raster_too_much_path, decimal=3
            )

    # Test encoding
    rast_1_xr = rasters.read(raster_path)
    rast_2_xr = rasters.read(raster2_path)
    rast_2_xr.encoding["add_field"] = True
    ci.assert_xr_encoding_attrs(rast_1_xr, rast_1_xr)
    with pytest.raises(AssertionError):
        ci.assert_xr_encoding_attrs(rast_1_xr, rast_2_xr)


@s3_env
def test_dim():
    """Test CI functions"""
    # Dim file
    dim_path = rasters_path().joinpath("dim_file.dim")
    dim_path_only_coh = rasters_path().joinpath("dim_file_only_coh.dim")
    dim_path_spk_only_coh = rasters_path().joinpath("dim_file_Spk_only_coh.dim")
    dim_path_different_image = rasters_path().joinpath("dim_file_other.dim")

    ci.assert_dim_file_equal(dim_path, dim_path)

    # Test if error when different number of bands
    with pytest.raises(AssertionError):
        ci.assert_dim_file_equal(dim_path, dim_path_only_coh)

    # Test if error when completely different
    with pytest.raises(AssertionError):
        ci.assert_dim_file_equal(dim_path, dim_path_different_image)

    # Test if error with same name different bands
    with pytest.raises(AssertionError):
        ci.assert_dim_file_equal(dim_path_only_coh, dim_path_spk_only_coh)


@s3_env
def test_assert_xml():
    # XML
    xml_folder = files_path().joinpath("LM05_L1TP_200030_20121230_20200820_02_T2_CI")
    xml_path = xml_folder.joinpath("LM05_L1TP_200030_20121230_20200820_02_T2_MTL.xml")
    xml_bad_path = xml_folder.joinpath("false_xml.xml")

    if path.is_cloud_path(files_path()):
        xml_path = xml_path.fspath
        xml_bad_path = xml_bad_path.fspath

    xml_ok = etree.parse(str(xml_path)).getroot()
    xml_nok = etree.parse(str(xml_bad_path)).getroot()

    ci.assert_xml_equal(xml_ok, xml_ok)
    with pytest.raises(AssertionError):
        ci.assert_xml_equal(xml_ok, xml_nok)


@s3_env
def test_assert_html():
    # HTML
    html_path = files_path().joinpath("productPreview.html")
    html_bad_path = files_path().joinpath("false.html")

    with tempfile.TemporaryDirectory() as tmp_dir:
        if path.is_cloud_path(files_path()):
            html_path = html_path.download_to(tmp_dir)
            html_bad_path = html_bad_path.download_to(tmp_dir)

        html_ok = etree.parse(str(html_path)).getroot()
        html_nok = etree.parse(str(html_bad_path)).getroot()

        ci.assert_xml_equal(html_ok, html_ok)
        with pytest.raises(AssertionError):
            ci.assert_xml_equal(html_ok, html_nok)



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Raster tools

You can use this only if you have installed sertit[full] or sertit[rasters_rio]
"""

import contextlib
import logging
import os
import tempfile
from functools import wraps
from typing import Any, Callable, Optional, Union

import geopandas as gpd
import numpy as np
import xarray as xr
from shapely.geometry import Polygon

try:
    import rasterio
    from rasterio import MemoryFile, features, merge, warp
    from rasterio import mask as rio_mask
    from rasterio import shutil as rio_shutil
    from rasterio.enums import Resampling
    from rasterio.vrt import WarpedVRT
    from rasterio.windows import Window, from_bounds
except ModuleNotFoundError as ex:
    raise ModuleNotFoundError(
        "Please install 'rasterio' to use the 'rasters_rio' package."
    ) from ex

from sertit import (
    AnyPath,
    geometry,
    logs,
    misc,
    os_utils,
    path,
    perf,
    strings,
    vectors,
    xml,
)
from sertit.logs import SU_NAME
from sertit.types import AnyNumpyArray, AnyPathStrType, AnyPathType, AnyRasterType

np.seterr(divide="ignore", invalid="ignore")

MAX_CORES = os.cpu_count() - 2
PATH_ARR_DS = AnyRasterType
LOGGER = logging.getLogger(SU_NAME)

DEG_2_RAD = np.pi / 180


# NODATAs
UINT8_NODATA = 255
""" :code:`uint8` nodata """

INT8_NODATA = -128
""" :code:`int8` nodata """

UINT16_NODATA = 65535
""" :code:`uint16` nodata """

FLOAT_NODATA = -9999
""" :code:`float` nodata """


def get_nodata_value_from_dtype(dtype) -> float:
    """
    Get default nodata value from any given dtype.

    Args:
        dtype: Dtype for the wanted nodata. Best if numpy's dtype.

    Returns:
        float: Nodata value

    Examples:
        >>> rasters_rio.get_nodata_value_from_dtype("uint8")
        255

        >>> rasters_rio.get_nodata_value_from_dtype("uint16")
        65535

        >>> rasters_rio.get_nodata_value_from_dtype("int8")
        -128

        >>> rasters_rio.get_nodata_value_from_dtype("float")
        -9999
    """
    # Convert type to numpy if needed
    with contextlib.suppress(AttributeError, TypeError):
        dtype = getattr(np, dtype)

    if dtype == np.uint8:
        nodata = UINT8_NODATA
    elif dtype == np.int8:
        nodata = INT8_NODATA
    elif dtype in [np.uint16, np.uint32, np.int32, np.int64, np.uint64, int, "int"]:
        nodata = UINT16_NODATA
    elif dtype in [np.int16, np.float32, np.float64, float, "float"]:
        nodata = FLOAT_NODATA
    else:
        LOGGER.warning(f"Not recognized dtype: {dtype}. Setting -9999 by default.")
        nodata = FLOAT_NODATA

    return nodata


def get_nodata_value(dtype) -> float:
    """
    .. deprecated:: 1.41.0
       Use :code:`get_nodata_value_from_dtype` instead.

    Get default nodata value:

    Args:
        dtype: Dtype for the wanted nodata. Best if numpy's dtype.

    Returns:
        float: Nodata value
    """
    logs.deprecation_warning(
        "This function is deprecated. Use 'get_nodata_value_from_dtype' instead."
    )
    return get_nodata_value_from_dtype(dtype)


def bigtiff_value(arr: Any) -> str:
    """
    Returns :code:`YES` if array is larger than 4 GB, :code:`IF_NEEDED` otherwise.

    Args:
        arr (Any): Numpy array or xarray

    Returns:
        str: YES or IF_NEEDED

    """
    try:
        itemsize = arr.itemsize
    except AttributeError:
        itemsize = arr.data.itemsize

    # Check size
    bigtiff = "YES" if arr.size * itemsize / 1024 / 1024 / 1024 > 4 else "IF_NEEDED"

    return bigtiff


def any_raster_to_rio_ds(function: Callable) -> Callable:
    """
    Allows a function to ingest AnyRasterType and convert it into a rasterio.DatasetReader:

    - a path (:code:`Path`, :code:`CloudPath` or :code:`str`)
    - a :code:`xarray.Dataset` or a :code:`xarray.DataArray`
    - a :code:`rasterio.DatasetWriter` or :code:`rasterio.DatasetReader`
    - :code:`rasterio` dataset after reading, its array and metadata: (np.ndarray, dict)

    Args:
        function (Callable): Function to decorate

    Returns:
        Callable: decorated function

    Example:
        >>> # Create mock function
        >>> @any_raster_to_rio_ds
        >>> def fct(ds):
        >>>     read(ds)
        >>>
        >>> # Test the two ways
        >>> read1 = fct("path/to/raster.tif")
        >>> with rasterio.open("path/to/raster.tif") as ds:
        >>>     read2 = fct(ds)
        >>>
        >>> # Test
        >>> read1 == read2
        True
    """

    @wraps(function)
    def wrapper(any_raster_type: AnyRasterType, *args, **kwargs) -> Any:
        """
        any_raster_to_rio_ds wrapper
        Args:
            any_raster_type (AnyRasterType): Raster path or its dataset
            *args: args
            **kwargs: kwargs

        Returns:
            Any: regular output
        """
        # Input is a path: open it with rasterio
        if path.is_path(any_raster_type):
            # GOTCHA: rasterio and cloudpathlib are not really compatible, so passing a CloudPath directly to rasterio (without turning it into a string) with cache the file!
            # This is really not ideal, so use the string conversion instead
            with rasterio.open(str(any_raster_type)) as ds:
                out = function(ds, *args, **kwargs)
        # Input is a tuple: we consider it's composed of an output of rasterio.read function, a numpy array and a metadata dict
        elif isinstance(any_raster_type, tuple):
            try:
                arr, meta = any_raster_type
                assert isinstance(arr, np.ndarray)
                assert isinstance(meta, dict)
            except (ValueError, AssertionError) as exc:
                raise TypeError(
                    "Input tuple should be composed of a numpy array of your data and the corresponding metadata dictionary, is rasterio's sense."
                ) from exc

            with (
                MemoryFile() as memfile,
                memfile.open(**meta, BIGTIFF=bigtiff_value(arr)) as ds,
            ):
                ds.write(arr)
                out = function(ds, *args, **kwargs)

        # Return given xarray object as is
        elif isinstance(any_raster_type, (xr.DataArray, xr.Dataset)):
            from sertit.rasters import get_nodata_value_from_xr

            nodata = get_nodata_value_from_xr(any_raster_type)

            meta = {
                "driver": "GTiff",
                "dtype": any_raster_type.dtype,
                "nodata": nodata,
                "width": any_raster_type.rio.width,
                "height": any_raster_type.rio.height,
                "count": any_raster_type.rio.count,
                "crs": any_raster_type.rio.crs,
                "transform": any_raster_type.rio.transform(),
            }
            with (
                MemoryFile() as memfile,
                memfile.open(**meta, BIGTIFF=bigtiff_value(any_raster_type)) as ds,
            ):
                if nodata is not None:
                    arr = any_raster_type.fillna(nodata)
                else:
                    arr = any_raster_type
                # /!\ Warning: this triggers a dask compute!
                ds.write(arr.data)
                out = function(ds, *args, **kwargs)

        # Run the fct directly on the input (which should be a rasterio Dataset). If not, this will fail and it's expected.
        else:
            out = function(any_raster_type, *args, **kwargs)

        return out

    return wrapper


def path_arr_dst(function: Callable) -> Callable:
    """
    .. deprecated:: 1.40.0
       Use :py:func:`rasters.any_raster_to_rio_ds` instead.
    """
    logs.deprecation_warning(
        "Deprecated 'path_arr_dst' decorator. Please use 'any_raster_to_rio_ds' instead."
    )
    return any_raster_to_rio_ds(function)


@any_raster_to_rio_ds
def get_new_shape(
    ds: AnyRasterType,
    resolution: Union[tuple, list, float],
    size: Union[tuple, list],
    window: Window = None,
) -> (int, int, bool):
    """
    Get the new shape (height, width) of a resampled raster.

    Size overrides resolution

    Args:
        ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata
        resolution (Union[tuple, list, float]): Resolution of the wanted band, in dataset resolution unit (X, Y)
        size (Union[tuple, list]): Size of the array (width, height). Not used if resolution is provided.
        window (Window): Window to be read

    Returns:
        (int, int, bool): Height, width, do resampling

    """
    # By default, do a resampling
    do_resampling = True

    def _get_new_dim(dim: int, res_old: float, res_new: float) -> (int, bool):
        """
        Get the new dimension in pixels
        Args:
            dim (int): Old dimension
            res_old (float): Old resolution
            res_new (float): New resolution

        Returns:
            (int, bool): New dimension, do resampling
        """
        new_dim = int(np.round(dim * res_old / res_new))
        do_res = dim != new_dim
        return new_dim, do_res

    # By default keep original shape
    if window is None:
        new_height = ds.height
        new_width = ds.width
    else:
        new_height = window.height
        new_width = window.width

    # Compute new shape
    if size is not None:
        try:
            if new_height == size[1] and new_width == size[0]:
                do_resampling = False
            else:
                new_height = size[1]
                new_width = size[0]
        except (TypeError, KeyError) as exc:
            raise ValueError(
                f"Size should be None or a castable to a list: {size}"
            ) from exc
    elif resolution is not None:
        if isinstance(resolution, (int, float)):
            new_height, do_resampling = _get_new_dim(new_height, ds.res[1], resolution)
            new_width, do_resampling = _get_new_dim(new_width, ds.res[0], resolution)
        else:
            try:
                if len(resolution) != 2:
                    raise ValueError(
                        "We should have a resolution for X and Y dimensions"
                    )

                if resolution[0] is not None:
                    new_width, do_resampling = _get_new_dim(
                        new_width, ds.res[0], resolution[0]
                    )

                if resolution[1] is not None:
                    new_height, do_resampling = _get_new_dim(
                        new_height, ds.res[1], resolution[1]
                    )
            except (TypeError, KeyError) as exc:
                raise ValueError(
                    f"Resolution should be None, 2 floats or a castable to a list: {resolution}"
                ) from exc
    else:
        do_resampling = False

    return new_height, new_width, do_resampling


def update_meta(arr: AnyNumpyArray, meta: dict) -> dict:
    """
    Basic metadata update from a numpy array. Updates everything that we can find in the array:

    - :code:`dtype`: array dtype,
    - :code:`count`: first dimension of the array if the array is in 3D, else 1
    - :code:height`: second dimension of the array
    - :code:`width`: third dimension of the array
    - :code:`nodata`: if a masked array is given, nodata is its fill_value

    .. WARNING::
        The array's shape is interpreted in rasterio's way (count, height, width) !

    Args:
        arr (AnyNumpyArray): Array from which to update the metadata
        meta (dict): Metadata to update

    Returns:
        dict: Update metadata

    Example:
        >>> raster_path = "path/to/raster.tif"
        >>> with rasterio.open(raster_path) as ds:
        >>>      meta = ds.meta
        >>>      arr = ds.read()
        >>> meta
        {
            'driver': 'GTiff',
            'dtype': 'float32',
            'nodata': None,
            'width': 300,
            'height': 300,
            'count': 4,
            'crs': CRS.from_epsg(32630),
            'transform': Affine(20.0, 0.0, 630000.0,0.0, -20.0, 4870020.0)
        }
        >>> new_arr = np.ma.masked_array(arr[:, ::2, ::2].astype(np.uint8), fill_value=0)
        >>> new_arr.shape
        (4, 150, 150)
        >>> new_arr.dtype
        dtype('uint8')
        >>> new_arr.fill_value
        0
        >>> update_meta(new_arr, meta)
        {
            'driver': 'GTiff',
            'dtype': dtype('uint8'),
            'nodata': 0,
            'width': 150,
            'height': 150,
            'count': 4,
            'crs': CRS.from_epsg(32630),
            'transform': Affine(20.0, 0.0, 630000.0, 0.0, -20.0, 4870020.0)
        }

    """
    # Manage raster shape (Stored in rasterio's way)
    shape = arr.shape
    count = 1 if len(shape) == 2 else shape[0]
    width = shape[-1]
    height = shape[-2]

    # Update metadata that can be derived from raster
    out_meta = meta.copy()
    out_meta.update(
        {"dtype": arr.dtype, "count": count, "height": height, "width": width}
    )

    # Nodata
    if isinstance(arr, np.ma.masked_array):
        out_meta["nodata"] = arr.fill_value

    return out_meta


def get_nodata_mask(
    array: AnyNumpyArray,
    has_nodata: bool,
    default_nodata: int = 0,
) -> np.ndarray:
    """
    .. deprecated:: 1.36.0
       Use :py:func:`rasters_rio.get_data_mask` instead.
    """
    logs.deprecation_warning("This function is deprecated. Use 'get_data_mask' instead")
    return get_data_mask(array, has_nodata, default_nodata)


def get_data_mask(
    array: AnyNumpyArray,
    has_nodata: bool,
    default_nodata: int = 0,
) -> np.ndarray:
    """
    Get nodata mask from a masked array.

    The nodata may not be set before, then pass a nodata value that will be evaluated on the array.

    .. WARNING::
        Sets 1 where the data is valid and 0 where it is not!

    Example:
        >>> diag_arr = np.diag([1,2,3])
        array([[1, 0, 0],
               [0, 2, 0],
               [0, 0, 3]])
        >>>
        >>> get_data_mask(diag_arr, has_nodata=False)
        array([[1, 0, 0],
               [0, 1, 0],
               [0, 0, 1]], dtype=uint8)
        >>>
        >>> get_data_mask(diag_arr, has_nodata=False, default_nodata=1)
        array([[0, 1, 1],
               [1, 1, 1],
               [1, 1, 1]], dtype=uint8)

    Args:
        array (np.ma.masked_array): Array to evaluate
        has_nodata (bool): If the array as its nodata specified. If not, using default_nodata.
        default_nodata (int): Default nodata used if the array's nodata is not set

    Returns:
        np.ndarray: Pixelwise nodata array

    """
    # Nodata mask
    if not has_nodata or not isinstance(
        array, np.ma.masked_array
    ):  # Unspecified nodata is set to None by rasterio
        if np.isnan(default_nodata):
            nodata_mask = np.where(np.isnan(array), 0, 1).astype(np.uint8)
        else:
            nodata_mask = np.where(array != default_nodata, 1, 0).astype(np.uint8)
    else:
        nodata_mask = np.where(array.mask, 0, 1).astype(np.uint8)

    return nodata_mask


@any_raster_to_rio_ds
def rasterize(
    ds: AnyRasterType,
    vector: Union[gpd.GeoDataFrame, AnyPathStrType],
    value_field: str = None,
    default_nodata: int = 0,
    default_value: int = 1,
    **kwargs,
) -> (np.ma.masked_array, dict):
    """
    Rasterize a vector into raster format.

    Note that passing `merge_alg = MergeAlg.add` will add the vector values to the given a raster

    See: https://pygis.io/docs/e_raster_rasterize.html

    Args:
        ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata
        vector (Union[gpd.GeoDataFrame, AnyPathStrType]): Vector to be rasterized
        value_field (str): Field of the vector with the values to be burnt on the raster (should be scalars). If let to None, the raster will be binary (`default_nodata`, `default_value`).
        default_nodata (int): Default nodata of the raster (outside the vector in the raster extent)
        default_value (int): Used as value for all geometries, if `value_field` not provided

    Returns:
        np.ma.masked_array, dict: Rasterized vector and its metadata
    """
    if not isinstance(vector, gpd.GeoDataFrame):
        vector = vectors.read(vector, crs=ds.crs)
    else:
        vector = vector.to_crs(crs=ds.crs)

    # Manage vector values
    if value_field:
        geom_value = (
            (geom, value) for geom, value in zip(vector.geometry, vector[value_field])
        )
        dtype = kwargs.pop("dtype", vector[value_field].dtype)
    else:
        geom_value = vector.geometry
        dtype = kwargs.pop("dtype", np.uint8)

    # Manage nodata
    if "nodata" in kwargs:
        nodata = kwargs.pop("nodata")
    elif ds.nodata:
        nodata = ds.nodata
    else:
        nodata = default_nodata

    # Check if the nodata value can be cast into the new dtype
    is_castable = np.can_cast(np.array(nodata, dtype=ds.dtypes[0]), dtype)

    # Floating point values that can be converted to integers are allowed (if they respect min / max for the specific dtype)
    min_max_dtype = np.iinfo(dtype)
    is_valid_int = (
        abs(nodata - int(nodata)) == 0
        and min_max_dtype.min < nodata < min_max_dtype.max
    )

    if not is_castable and not is_valid_int:
        old_nodata = nodata
        nodata = get_nodata_value_from_dtype(dtype)

        # Only throw a warning if the value is really different  (we don't care about 255.0 being replaced by 255)
        if old_nodata - nodata != 0.0:
            LOGGER.warning(
                f"Impossible to cast nodata value ({old_nodata}) into the wanted dtype ({str(dtype)}). "
                f"Default nodata value for this current dtype will be used ({nodata})."
            )

    # Rasterize vector
    mask = features.rasterize(
        geom_value,
        out_shape=(ds.height, ds.width),
        fill=nodata,  # Outside vector
        default_value=default_value,  # Inside vector
        transform=ds.transform,
        dtype=dtype,
        all_touched=kwargs.get("all_touched", True),
        **misc.select_dict(kwargs, ["merge_alg"]),
    )

    meta = ds.meta.copy()
    meta["dtype"] = dtype
    meta["nodata"] = nodata

    return mask, meta


@any_raster_to_rio_ds
def _vectorize(
    ds: AnyRasterType,
    values: Union[None, int, list] = None,
    keep_values: bool = True,
    dissolve: bool = False,
    get_nodata: bool = False,
    default_nodata: int = 0,
) -> gpd.GeoDataFrame:
    """
    Vectorize a raster, both to get classes or nodata.

    If dissolved is False, it returns a GeoDataFrame with a GeoSeries per cluster of pixel value,
    with the value as an attribute. Else it returns a GeoDataFrame with a unique polygon.

    .. WARNING::
        If :code:`get_nodata` is set to False:
            - Please only use this function on a classified raster.
            - This could take a while as the computing time directly depends on the number of polygons to vectorize.
                Please be careful.
        Else:
            - You will get a classified polygon with data (value=0)/nodata pixels.

    Args:
        ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata
        values (Union[None, int, list]): Get only the polygons concerning this/these particular values
        keep_values (bool): Keep the passed values. If False, discard them and keep the others.
        dissolve (bool): Dissolve all the polygons into one unique. Only works if values are given.
        get_nodata (bool): Get nodata vector (raster values are set to 0, nodata values are the other ones)
        default_nodata (int): Default values for nodata in case of non existing in file
    Returns:
        gpd.GeoDataFrame: Vector
    """
    # Get the shapes
    array = ds.read(masked=True)

    # Manage nodata value
    has_nodata = ds.nodata is not None
    nodata = ds.nodata if has_nodata else default_nodata

    # Manage values
    if values is not None:
        if not isinstance(values, list):
            values = [values]

        # If we want a dissolved vector, just set 1instead of real values
        arr_vals = 1 if dissolve else array
        if keep_values:
            true = arr_vals
            false = nodata
        else:
            true = nodata
            false = arr_vals

        data = np.where(np.isin(array, values), true, false).astype(array.dtype)
    else:
        data = array.data

    # Get nodata mask
    nodata_mask = get_data_mask(data, has_nodata=False, default_nodata=nodata)

    # Get shapes (on array or on mask to get nodata vector)
    shapes = features.shapes(
        nodata_mask if get_nodata else data,
        mask=None if get_nodata else nodata_mask,
        transform=ds.transform,
    )

    # Convert to geodataframe (with valid geometries)
    gdf = vectors.shapes_to_gdf(shapes, ds.crs)

    # Dissolve if needed
    if dissolve:
        gdf = gpd.GeoDataFrame(geometry=gdf.geometry, crs=gdf.crs).dissolve()

    return gdf


@any_raster_to_rio_ds
def vectorize(
    ds: AnyRasterType,
    values: Union[None, int, list] = None,
    keep_values: bool = True,
    dissolve: bool = False,
    default_nodata: int = 0,
) -> gpd.GeoDataFrame:
    """
    Vectorize a raster to get the class vectors.

    If dissolved is False, it returns a GeoDataFrame with a GeoSeries per cluster of pixel value,
    with the value as an attribute. Else it returns a GeoDataFrame with a unique polygon.

    .. WARNING::
        - Please only use this function on a classified raster.
        - This could take a while as the computing time directly depends on the number of polygons to vectorize.
            Please be careful.

    Args:
        ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata
        values (Union[None, int, list]): Get only the polygons concerning this/these particular values
        keep_values (bool): Keep the passed values. If False, discard them and keep the others.
        dissolve (bool): Dissolve all the polygons into one unique. Only works if values are given.
        default_nodata (int): Default values for nodata in case of non-existing in file

    Returns:
        gpd.GeoDataFrame: Classes Vector

    Example:
        >>> raster_path = "path/to/raster.tif"  # Classified raster, with no data set to 255
        >>> vec1 = vectorize(raster_path)
        >>> # or
        >>> with rasterio.open(raster_path) as ds:
        >>>     vec2 = vectorize(ds)
        >>>
        >>> # Assert those two approaches give the same result
        >>> vec1 == vec2
        True
    """
    return _vectorize(
        ds,
        values=values,
        get_nodata=False,
        keep_values=keep_values,
        dissolve=dissolve,
        default_nodata=default_nodata,
    )


@any_raster_to_rio_ds
def get_valid_vector(ds: AnyRasterType, default_nodata: int = 0) -> gpd.GeoDataFrame:
    """
    Get the valid data of a raster as a vector.

    Pay attention that every nodata pixel will appear too.
    If you want only the footprint of the raster, please use :code:`get_footprint`.

    Args:
        ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata
        default_nodata (int): Default values for nodata in case of non-existing in file

    Returns:
        gpd.GeoDataFrame: Nodata Vector

    Example:
        >>> raster_path = "path/to/raster.tif"  # Classified raster, with no data set to 255
        >>> nodata1 = get_nodata_vec(raster_path)
        >>>
        >>> # or
        >>> with rasterio.open(raster_path) as ds:
        >>>     nodata2 = get_nodata_vec(ds)
        >>>
        >>> # Assert those two approaches give the same result
        >>> nodata1 == nodata2
        True

    """
    nodata = _vectorize(ds, values=None, get_nodata=True, default_nodata=default_nodata)
    return nodata[
        nodata.raster_val != 0
    ]  # 0 is the values of not nodata put there by rasterio


@any_raster_to_rio_ds
def get_nodata_vector(ds: AnyRasterType, default_nodata: int = 0) -> gpd.GeoDataFrame:
    """
    Get the nodata vector of a raster as a vector.

    Pay attention that every nodata pixel will appear too.
    If you want only the footprint of the raster, please use :code:`get_footprint`.

    Args:
        ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata
        default_nodata (int): Default values for nodata in case of non-existing in file

    Returns:
        gpd.GeoDataFrame: Nodata Vector

    Example:
        >>> raster_path = "path/to/raster.tif"  # Classified raster, with no data set to 255
        >>> nodata1 = get_nodata_vec(raster_path)
        >>> # or
        >>> with rasterio.open(raster_path) as ds:
        >>>     nodata2 = get_nodata_vec(ds)
        >>>
        >>> # Assert those two approaches give the same result
        >>> nodata1 == nodata2
        True

    """
    nodata = _vectorize(ds, values=None, get_nodata=True, default_nodata=default_nodata)
    return nodata[
        nodata.raster_val == 0
    ]  # 0 is the values of not nodata put there by rasterio


@any_raster_to_rio_ds
def _mask(
    ds: AnyRasterType,
    shapes: Union[gpd.GeoDataFrame, Polygon, list],
    nodata: Optional[int] = None,
    do_crop: bool = False,
    **kwargs,
) -> (np.ma.masked_array, dict):
    """
    Overload of rasterio mask function in order to create a masked_array.

    The :code:`mask` function docs can be seen `here <https://rasterio.readthedocs.io/en/latest/api/rasterio.mask.html>`_.

    It basically masks a raster with a vector mask, with the possibility to crop the raster to the vector's extent.

    Args:
        ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata
        shapes (Union[gpd.GeoDataFrame, Polygon, list]): Shapes with the same CRS as the dataset
            (except if a :code:`GeoDataFrame` is passed, in which case it will automatically be converted.
        nodata (int): Nodata value. If not set, uses the ds.nodata. If doesnt exist, set to 0.
        do_crop (bool): Whether to crop the raster to the extent of the shapes. Default is False.
        **kwargs: Other rasterio.mask options

    Returns:
         (np.ma.masked_array, dict): Cropped array as a masked array and its metadata
    """
    if isinstance(shapes, (gpd.GeoDataFrame, gpd.GeoSeries)):
        shapes = shapes.to_crs(ds.crs).geometry
    elif not isinstance(shapes, list):
        shapes = [shapes]

    # Set nodata
    if not nodata:
        nodata = ds.nodata if ds.nodata else 0

    # Crop dataset
    possible_kwargs = ["all_touched", "invert", "filled", "pad", "pad_width", "indexes"]
    msk, trf = rio_mask.mask(
        ds,
        shapes,
        nodata=nodata,
        crop=do_crop,
        **misc.select_dict(kwargs, possible_kwargs),
    )

    # Create masked array
    nodata_mask = np.where(msk == nodata, 1, 0).astype(np.uint8)
    mask_array = np.ma.masked_array(msk, nodata_mask, fill_value=nodata)

    # Update meta
    out_meta = update_meta(mask_array, ds.meta)
    out_meta["transform"] = trf

    return mask_array, out_meta


@any_raster_to_rio_ds
def mask(
    ds: AnyRasterType,
    shapes: Union[gpd.GeoDataFrame, Polygon, list],
    nodata: Optional[int] = None,
    **kwargs,
) -> (np.ma.masked_array, dict):
    """
    Masking a dataset:
    setting nodata outside of the given shapes, but without cropping the raster to the shapes extent.

    Overload of rasterio mask function in order to create a masked_array.

    The :code:`mask` function docs can be seen `here <https://rasterio.readthedocs.io/en/latest/api/rasterio.mask.html>`_.
    It basically masks a raster with a vector mask, with the possibility to crop the raster to the vector's extent.

    Args:
        ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata
        shapes (Union[gpd.GeoDataFrame, Polygon, list]): Shapes with the same CRS as the dataset
            (except if a :code:`GeoDataFrame` is passed, in which case it will automatically be converted.
        nodata (int): Nodata value. If not set, uses the ds.nodata. If doesnt exist, set to 0.
        **kwargs: Other rasterio.mask options

    Returns:
         (np.ma.masked_array, dict): Masked array as a masked array and its metadata

    Example:
        >>> raster_path = "path/to/raster.tif"
        >>> shape_path = "path/to/shapes.geojson"  # Any vector that geopandas can read
        >>> shapes = gpd.read_file(shape_path)
        >>> masked_raster1, meta1 = mask(raster_path, shapes)
        >>> # or
        >>> with rasterio.open(raster_path) as ds:
        >>>     masked_raster2, meta2 = mask(ds, shapes)
        >>>
        >>> # Assert those two approaches give the same result
        >>> masked_raster1 == masked_raster2
        True
        >>> meta1 == meta2
        True
    """
    return _mask(ds, shapes=shapes, nodata=nodata, do_crop=False, **kwargs)


@any_raster_to_rio_ds
def crop(
    ds: AnyRasterType,
    shapes: Union[gpd.GeoDataFrame, Polygon, list],
    nodata: Optional[int] = None,
    **kwargs,
) -> (np.ma.masked_array, dict):
    """
    Cropping a dataset:
    setting nodata outside of the given shapes AND cropping the raster to the shapes extent.

    **HOW:**

    Overload of rasterio mask function in order to create a masked_array.

    The :code:`mask` function docs can be seen `here <https://rasterio.readthedocs.io/en/latest/api/rasterio.mask.html>`_.
    It basically masks a raster with a vector mask, with the possibility to crop the raster to the vector's extent.

    Args:
        ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata
        shapes (Union[gpd.GeoDataFrame, Polygon, list]): Shapes with the same CRS as the dataset
            (except if a :code:`GeoDataFrame` is passed, in which case it will automatically be converted.
        nodata (int): Nodata value. If not set, uses the ds.nodata. If doesnt exist, set to 0.
        **kwargs: Other rasterio.mask options

    Returns:
         (np.ma.masked_array, dict): Cropped array as a masked array and its metadata

    Example:
        >>> raster_path = "path/to/raster.tif"
        >>> shape_path = "path/to/shapes.geojson"  # Any vector that geopandas can read
        >>> shapes = gpd.read_file(shape_path)
        >>> cropped_raster1, meta1 = crop(raster_path, shapes)
        >>> # or
        >>> with rasterio.open(raster_path) as ds:
        >>>     cropped_raster2, meta2 = crop(ds, shapes)
        >>>
        >>> # Assert those two approaches give the same result
        >>> cropped_raster1 == cropped_raster2
        True
        >>> meta1 == meta2
        True
    """
    return _mask(ds, shapes=shapes, nodata=nodata, do_crop=True, **kwargs)


@any_raster_to_rio_ds
def get_window(ds: AnyRasterType, window: Any):
    """
    Get a window from any type of input

    Args:
        ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata
        window (Any): Anything that can be returned as a window. In case of iterable, assumption is made it's geographic bounds. For pixel, please provide a Window directly.

    Returns:
        Window: Rasterio window

    """
    if window is not None and not isinstance(window, Window):
        if isinstance(window, gpd.GeoDataFrame):
            bounds = window.to_crs(ds.crs).total_bounds
        elif path.is_path(window):
            bounds = vectors.read(window).to_crs(ds.crs).total_bounds
        else:
            bounds = window

        try:
            window = from_bounds(*bounds, ds.transform)
        except Exception as exc:
            raise TypeError(
                "Window should either be a GeoDataFrame, tuple, list, Window, readable as a vector or set to None"
            ) from exc

    # Use rioxarray way to convert window to integer
    (row_start, row_stop), (col_start, col_stop) = window.toranges()
    row_start = 0 if row_start < 0 else np.floor(row_start)
    row_stop = 0 if row_stop < 0 else np.ceil(row_stop)
    col_start = 0 if col_start < 0 else np.floor(col_start)
    col_stop = 0 if col_stop < 0 else np.ceil(col_stop)
    row_slice = slice(int(row_start), int(row_stop))
    col_slice = slice(int(col_start), int(col_stop))

    window = Window.from_slices(
        rows=row_slice,
        cols=col_slice,
        width=int(window.width),
        height=int(window.height),
    )
    return window


@any_raster_to_rio_ds
def read(
    ds: AnyRasterType,
    resolution: Union[tuple, list, float] = None,
    size: Union[tuple, list] = None,
    window: Any = None,
    resampling: Resampling = Resampling.nearest,
    masked: bool = True,
    **kwargs,
) -> (np.ma.masked_array, dict):
    """
    Read a raster dataset from a :code:`rasterio.Dataset` or a path.

    The resolution can be provided (in dataset unit) as:

    - a tuple or a list of (X, Y) resolutions
    - a float, in which case X resolution = Y resolution
    - None, in which case the dataset resolution will be used

    Tip:
    Use index with a list of one element to keep a 3D array

    Args:
        ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata
        resolution (Union[tuple, list, float]): Resolution of the wanted band, in dataset resolution unit (X, Y)
        size (Union[tuple, list]): Size of the array (width, height). Not used if resolution is provided.
        window (Any): Anything that can be returned as a window (i.e. path, gpd.GeoPandas, Iterable, rasterio.Window...).
            In case of an iterable, assumption is made it corresponds to geographic bounds.
            For pixel, please provide a rasterio.Window directly.
        resampling (Resampling): Resampling method (nearest by default)
        masked (bool): Get a masked array, :code:`True` by default (whereas it is False by default in rasterio)
        **kwargs: Other ds.read() arguments such as indexes.

    Returns:
        np.ma.masked_array, dict: Masked array corresponding to the raster data and its metadata

    Example:
        >>> raster_path = "path/to/raster.tif"
        >>> raster1, meta1 = read(raster_path)
        >>> # or
        >>> with rasterio.open(raster_path) as ds:
        >>>    raster2, meta2 = read(ds)
        >>>
        >>> # Assert those two approaches give the same result
        >>> raster1 == raster2
        True
        >>> meta1 == meta2
        True

    """
    dst_transform = None
    if window is not None:
        window = get_window(ds, window)
        dst_transform = ds.window_transform(window)

    # Get new height and width
    new_height, new_width, _ = get_new_shape(ds, resolution, size, window)

    # Manage out_shape
    if "indexes" in kwargs:
        if isinstance(kwargs["indexes"], int):
            out_shape = (new_height, new_width)
            new_count = 1
        else:
            new_count = len(kwargs["indexes"])
            out_shape = (new_count, new_height, new_width)
    else:
        new_count = ds.count
        out_shape = (new_count, new_height, new_width)

    # Read data
    array = ds.read(
        out_shape=out_shape,
        resampling=resampling,
        masked=masked,
        window=window,
        **misc.select_dict(kwargs, ["indexes", "out_dtype", "boundless", "fill_value"]),
    )

    # Get destination transform
    if dst_transform is None:
        dst_transform = ds.transform * ds.transform.scale(
            (ds.width / new_width), (ds.height / new_height)
        )
    if window is not None and resolution:
        dst_transform = dst_transform * dst_transform.scale(
            (window.width / new_width), (window.height / new_height)
        )

    # Update meta
    dst_meta = ds.meta.copy()
    dst_meta.update(
        {
            "height": new_height,
            "width": new_width,
            "count": new_count,
            "transform": dst_transform,
            "dtype": array.dtype,
            "nodata": ds.nodata,
        }
    )

    return array, dst_meta


def write(
    raster: AnyRasterType,
    meta: dict,
    output_path: AnyPathStrType = None,
    tags: dict = None,
    **kwargs,
) -> None:
    """
    Write raster to disk (encapsulation of rasterio's function)

    Metadata will be copied and updated with raster's information (i.e. width, height, count, type...)
    The driver is GTiff by default, and no nodata value is provided.
    The file will be compressed if the raster is a mask (saved as uint8)

    Args:
        raster (AnyRasterType): Raster to save on disk
        output_path (AnyPathStrType): Path where to save it (directories should be existing)
        tags (dict): Tags to write to the GeoTiff
        **kwargs: Overloading metadata, ie :code:`nodata=255`

    Example:
        >>> raster_path = "path/to/raster.tif"
        >>> raster_out = "path/to/out.tif"
        >>>
        >>> # Read raster
        >>> raster, meta = read(raster_path)
        >>>
        >>> # Rewrite it on disk
        >>> write(raster, meta, raster_out)
    """
    if output_path is None:
        logs.deprecation_warning(
            "'path' is deprecated in 'rasters_rio.write'. Use 'output_path' instead."
        )
        output_path = kwargs.pop("path")

    raster_out = raster.copy()

    # Prune empty kwargs to avoid throwing GDAL warnings/errors
    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    # Manage raster type (impossible to write boolean arrays)
    if raster_out.dtype == bool:
        raster_out = raster_out.astype(np.uint8)

    # Update metadata
    out_meta = meta.copy()

    # Update raster to be sure to write down correct nodata pixels
    nodata = kwargs.get("nodata")
    if nodata is None:
        nodata = meta.get("nodata")
        if nodata is None and isinstance(raster_out, np.ma.masked_array):
            nodata = raster_out.fill_value

    # TODO: change this with rasterio 1.3.0 (masked option in write)
    if isinstance(raster_out, np.ma.masked_array):
        raster_out[raster_out.mask] = nodata

    out_meta["nodata"] = nodata

    # Force compression and driver (but can be overwritten by kwargs)
    out_meta["driver"] = kwargs.get("driver", "GTiff")

    # Compress to LZW by default
    out_meta["compress"] = kwargs.get("compress", "lzw")

    if (
        out_meta["compress"].lower() in ["lzw", "deflate", "zstd"]
        and "predictor" not in out_meta  # noqa: W503
    ):
        if out_meta["dtype"] in [np.float32, np.float64, float]:
            out_meta["predictor"] = "3"
        else:
            out_meta["predictor"] = "2"

    # Bigtiff if needed (more than 4Go)
    out_meta["BIGTIFF"] = bigtiff_value(raster_out)

    # Set more threads
    out_meta["NUM_THREADS"] = perf.get_max_cores()

    # Update metadata with array data
    out_meta = update_meta(raster_out, out_meta)

    # Update metadata with additional params
    for key, val in kwargs.items():
        out_meta[key] = val

    # Manage raster shape
    if len(raster_out.shape) == 2:
        raster_out = np.expand_dims(raster_out, axis=0)

    # Write product
    with rasterio.open(str(output_path), "w", **out_meta) as ds:
        ds.write(raster_out)
        if tags is not None:
            ds.update_tags(**tags)


def collocate(
    reference_meta: dict,
    other_arr: AnyNumpyArray,
    other_meta: dict,
    resampling: Resampling = Resampling.nearest,
) -> (AnyNumpyArray, dict):
    """
    Collocate two georeferenced arrays:
    forces the *other* raster to be exactly georeferenced onto the *reference* raster by reprojection.

    Args:
        reference_meta (dict): Reference metadata
        other_arr (np.ma.masked_array): Other array to be collocated
        other_meta (dict): Other metadata
        resampling (Resampling): Resampling method

    Returns:
        np.ma.masked_array, dict: Collocated array and its metadata

    Example:
        >>> reference_path = "path/to/reference.tif"
        >>> other_path = "path/to/other.tif"
        >>> col_path = "path/to/collocated.tif"
        >>>
        >>> # Just open the master data
        >>> with rasterio.open(reference_path) as reference_dst:
        >>>     # Read other
        >>>     other, other_meta = read(other_path)
        >>>
        >>>     # Collocate the other to the reference
        >>>     col_arr, col_meta = collocate(reference_dst.meta,
        >>>                                   other,
        >>>                                   other_meta,
        >>>                                   Resampling.bilinear)
        >>>
        >>> # Write it
        >>> write(col_arr, col_path, col_meta)

    """
    collocated_arr = np.zeros(
        (reference_meta["count"], reference_meta["height"], reference_meta["width"]),
        dtype=reference_meta["dtype"],
    )
    warp.reproject(
        source=other_arr,
        destination=collocated_arr,
        src_transform=other_meta["transform"],
        src_crs=other_meta["crs"],
        dst_transform=reference_meta["transform"],
        dst_crs=reference_meta["crs"],
        src_nodata=other_meta["nodata"],
        dst_nodata=other_meta["nodata"],
        resampling=resampling,
        num_threads=perf.get_max_cores(),
    )

    meta = reference_meta.copy()
    meta.update(
        {
            "dtype": other_meta["dtype"],
            "driver": other_meta["driver"],
            "nodata": other_meta["nodata"],
        }
    )

    if isinstance(other_arr, np.ma.masked_array):
        collocated_arr = np.ma.masked_array(
            collocated_arr, other_arr.mask, fill_value=other_meta["nodata"]
        )

    return collocated_arr, meta


def sieve(
    array: AnyNumpyArray,
    out_meta: dict,
    sieve_thresh: int,
    connectivity: int = 4,
) -> (AnyNumpyArray, dict):
    """
    Sieving, overloads rasterio function with raster shaped like (1, h, w).

    Forces the output to :code:`np.uint8` (as only classified rasters should be sieved)

    Args:
        array (AnyNumpyArray): Array to sieve
        out_meta (dict): Metadata to update
        sieve_thresh (int): Sieving threshold in pixels
        connectivity (int): Connectivity, either 4 or 8

    Returns:
        (AnyNumpyArray, dict): Sieved array and updated meta

    Example:
        >>> raster_path = "path/to/raster.tif"  # classified raster
        >>>
        >>> # Read raster
        >>> raster, meta = read(raster_path)
        >>>
        >>> # Rewrite it
        >>> sieved, sieved_meta = sieve(raster, meta, sieve_thresh=20)
        >>>
        >>> # Write it
        >>> raster_out = "path/to/raster_sieved.tif"
        >>> write(sieved, raster_out, sieved_meta)
    """
    assert connectivity in [4, 8]

    # Read extraction array
    expand = False
    if len(array.shape) == 3 and array.shape[0] == 1:
        array = np.squeeze(array)  # Use this trick to make the sieve work
        expand = True

    # Get nodata mask
    msk = ~array.mask if isinstance(array, np.ma.masked_array) else np.ones_like(array)

    if expand:
        msk = np.squeeze(msk)

    # Convert to np.uint8 if needed
    dtype = np.uint8
    meta = out_meta.copy()
    if meta["dtype"] != dtype:
        array = array.astype(dtype)
        meta["dtype"] = dtype

    # Sieve
    result_array = np.empty(array.shape, dtype=array.dtype)
    features.sieve(
        array, size=sieve_thresh, out=result_array, connectivity=connectivity, mask=msk
    )

    # Use this trick to get the array back to 'normal'
    if expand:
        result_array = np.expand_dims(result_array, axis=0)

    return result_array, meta


def get_dim_img_path(
    dim_path: AnyPathStrType, img_name: str = "*", get_list: bool = False
) -> Union[list, AnyPathType]:
    """
    Get the image path (:code:`.img`) from a :code:`BEAM-DIMAP` data.

    A :code:`BEAM-DIMAP` file cannot be opened by rasterio, although its :code:`.img` file can.

    Args:
        dim_path (AnyPathStrType): DIM path (.dim or .data)
        img_name (str): .img file name (or regex), in case there are multiple .img files (i.e. for S3 data)

    Returns:
        AnyPathType: .img file

    Example:
        >>> dim_path = "path/to/dimap.dim"  # BEAM-DIMAP image
        >>> img_path = get_dim_img_path(dim_path)
        img_path = "path/to/dimap.dim/dimap.img"
        >>>
        >>> # Read raster
        >>> raster, meta = read(img_path)
    """
    dim_path = AnyPath(dim_path)
    if dim_path.suffix == ".dim":
        dim_path = dim_path.with_suffix(".data")

    assert dim_path.suffix == ".data" and dim_path.is_dir()

    return path.get_file_in_dir(
        dim_path, img_name, extension="img", exact_name=True, get_list=get_list
    )


@any_raster_to_rio_ds
def get_extent(ds: AnyRasterType) -> gpd.GeoDataFrame:
    """
    Get the extent of a raster as a :code:`geopandas.Geodataframe`.

    Args:
        ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata

    Returns:
        gpd.GeoDataFrame: Extent as a  :code:`geopandas.Geodataframe`

    Example:
        >>> raster_path = "path/to/raster.tif"
        >>>
        >>> extent1 = get_extent(raster_path)
        >>> # or
        >>> with rasterio.open(raster_path) as ds:
        >>>     extent2 = get_extent(ds)
        >>>
        >>> # Assert those two approaches give the same result
        >>> extent1 == extent2
        True
    """
    return vectors.get_geodf(geom=[*ds.bounds], crs=ds.crs)


@any_raster_to_rio_ds
def get_footprint(ds: AnyRasterType) -> gpd.GeoDataFrame:
    """
    Get real footprint of the product (without nodata, in *french == emprise utile*)

    Args:
        ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata

    Returns:
        gpd.GeoDataFrame: Footprint as a GeoDataFrame

    Example:
        >>> raster_path = "path/to/raster.tif"
        >>>
        >>> footprint1 = get_footprint(raster_path)
        >>>
        >>> # or
        >>> with rasterio.open(raster_path) as ds:
        >>>     footprint2 = get_footprint(ds)
        >>>
        >>> # Assert those two approaches give the same result
        >>> footprint1 == footprint2
    """
    footprint = get_valid_vector(ds)

    return geometry.get_wider_exterior(footprint)


def merge_vrt(
    paths: list,
    merged_path: AnyPathStrType,
    abs_path: bool = False,
    **kwargs,
) -> None:
    """
    Merge rasters as a VRT. Uses :code:`gdalbuildvrt`.

    See here: https://gdal.org/programs/gdalbuildvrt.html

    Creates VRT with relative paths!

    This function handles files of different projection by creating intermediate VRT used for warping.
    All VRTs will be written with relative paths.

    Args:
        paths (list): Path of the rasters to be merged with the same CRS)
        merged_path (AnyPathStrType): Path to the merged raster
        abs_path (bool): VRT with absolute paths. If not, VRT with relative paths (default)
        kwargs: Other :code:`gdalbuildvrt` arguments

    Example:
        >>> paths_utm32630 = ["path/to/raster1.tif", "path/to/raster2.tif", "path/to/raster3.tif"]
        >>> paths_utm32631 = ["path/to/raster4.tif", "path/to/raster5.tif"]
        >>>
        >>> mosaic_32630 = "path/to/mosaic_32630.vrt"
        >>> mosaic_32631 = "path/to/mosaic_32631.vrt"
        >>>
        >>> # Create mosaic, one by CRS !
        >>> merge_vrt(paths_utm32630, mosaic_32630)
        >>> merge_vrt(paths_utm32631, mosaic_32631, {"-srcnodata":255, "-vrtnodata":0})
    """
    # Copy crs_paths in order not to modify it in place (replacing str by Paths for example)
    crs_paths_cp = paths.copy()

    # Manage cloud paths (gdalbuildvrt needs url or true filepaths)
    merged_path = AnyPath(merged_path)

    first_crs = kwargs.get("crs")
    for i, crs_path in enumerate(crs_paths_cp):
        crs_path = AnyPath(crs_path)
        # Download file if VRT is needed
        if path.is_cloud_path(crs_path):
            crs_path = crs_path.download_to(merged_path.parent)

        with rasterio.open(str(crs_path)) as src:
            if first_crs is None:
                first_crs = src.crs
            else:
                # Reproject bands if needed
                if first_crs != src.crs:
                    crs_epsg = first_crs.to_epsg()
                    with WarpedVRT(src, **{"crs": first_crs.to_epsg()}) as vrt:
                        # At this point 'vrt' is a full dataset with dimensions,
                        # CRS, and spatial extent matching 'vrt_options'.
                        new_crs_name = path.get_filename(crs_path) + f"_{crs_epsg}.vrt"
                        new_crs_path = os.path.join(merged_path.parent, new_crs_name)
                        rio_shutil.copy(vrt, new_crs_path, driver="vrt")

                        try:
                            # Set to relative
                            # This is clearly a hack, but GDAL doesn't handle any copy with relative path
                            # See https://github.com/rasterio/rasterio/discussions/2720
                            vrt_root = xml.read(new_crs_path)
                            xml.update_attrib(
                                vrt_root, "SourceDataset", "relativeToVRT", "1"
                            )
                            xml.update_txt(
                                vrt_root,
                                "SourceDataset",
                                crs_path.relative_to(merged_path.parent),
                            )
                            xml.write(vrt_root, new_crs_path)
                        except ValueError as ex:
                            LOGGER.warning(f"Your VRT will be absolute as {str(ex)}")

                        # Add in place
                        crs_path = new_crs_path

        crs_paths_cp[i] = crs_path

    # Create relative paths
    vrt_root = os.path.dirname(merged_path)
    try:
        if abs_path:
            paths = [strings.to_cmd_string(path.to_abspath(p)) for p in crs_paths_cp]
            merged_path = strings.to_cmd_string(merged_path.resolve())
        else:
            paths = [
                strings.to_cmd_string(path.real_rel_path(p, vrt_root))
                for p in crs_paths_cp
            ]
            merged_path = strings.to_cmd_string(
                path.real_rel_path(merged_path, vrt_root)
            )

    except ValueError:
        # ValueError when crs_merged_path and crs_paths are not on the same disk
        paths = [strings.to_cmd_string(str(p)) for p in crs_paths_cp]
        merged_path = strings.to_cmd_string(str(merged_path))

    # Run cmd
    gdal_build_vrt_exe = os_utils.gdalbuildvrt_exe()
    arg_list = [val for item in kwargs.items() for val in item]
    try:
        vrt_cmd = [gdal_build_vrt_exe, merged_path, *paths, *arg_list]
        misc.run_cli(vrt_cmd, cwd=vrt_root)

    except RuntimeError:
        # Manage too long command line
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_file = os.path.join(tmp_dir, "list.txt")
            with open(tmp_file, "w+") as f:
                for p in paths:
                    p = p.replace('"', "")
                    f.write(f"{p}\n")

            vrt_cmd = [
                gdal_build_vrt_exe,
                "-input_file_list",
                tmp_file,
                merged_path,
                *arg_list,
            ]

            misc.run_cli(vrt_cmd, cwd=vrt_root)


def merge_gtiff(paths: list, merged_path: AnyPathStrType, **kwargs) -> None:
    """
    Merge rasters as a GeoTiff.

    This function handles files of different projection by creating intermediate VRT used for warping.

    Using :code:`rasterio.merge` with default arguments behind the hood so you can provide any valid argument into the kwargs.
    For example, to modify the merging method, you can pass :code:`method="max"`

    Args:
        paths (list): Path of the rasters to be merged with the same CRS)
        merged_path (AnyPathStrType): Path to the merged raster
        kwargs: Other rasterio.merge arguments
            More info `here <https://rasterio.readthedocs.io/en/latest/api/rasterio.merge.html#rasterio.merge.merge>`_

    Example:
        >>> paths_utm32630 = ["path/to/raster1.tif", "path/to/raster2.tif", "path/to/raster3.tif"]
        >>> paths_utm32631 = ["path/to/raster4.tif", "path/to/raster5.tif"]
        >>>
        >>> mosaic_32630 = "path/to/mosaic_32630.tif"
        >>> mosaic_32631 = "path/to/mosaic_32631.tif"
        >>>
        >>> # Create mosaic, one by CRS !
        >>> merge_gtiff(paths_utm32630, mosaic_32630)
        >>> merge_gtiff(paths_utm32631, mosaic_32631)
    """
    # Open datasets for merging
    tmp_dir = None
    crs_datasets = []
    try:
        first_crs = kwargs.get("crs")
        for tile_path in paths:
            src = rasterio.open(str(tile_path))
            if first_crs is None:
                first_crs = src.crs
            else:
                # Reproject bands if needed
                if first_crs != src.crs:
                    tmp_dir = tempfile.TemporaryDirectory()
                    with WarpedVRT(src, **{"crs": first_crs.to_epsg()}) as vrt:
                        # At this point 'vrt' is a full dataset with dimensions,
                        # CRS, and spatial extent matching 'vrt_options'.
                        tile_path = os.path.join(
                            tmp_dir.name, path.get_filename(tile_path) + ".vrt"
                        )
                        rio_shutil.copy(vrt, tile_path, driver="vrt")

                    src.close()
                    src = rasterio.open(str(tile_path))

            crs_datasets.append(src)

        # Merge all datasets
        merge_args = [
            "bounds",
            "res",
            "precision",
            "indexes",
            "output_count",
            "resampling",
            "method",
            "target_aligned_pixels",
            "mem_limit",
            "use_highest_res",
            "masked",
            "dst_path",
            "dst_kwds",
        ]
        merge_kwargs = misc.select_dict(kwargs, merge_args + ["nodata", "dtype"])
        write_kwargs = misc.prune_dict(kwargs, merge_args)

        merged_array, merged_transform = merge.merge(crs_datasets, **merge_kwargs)
        merged_meta = crs_datasets[0].meta.copy()
        merged_meta.update(
            {
                "driver": "GTiff",
                "height": merged_array.shape[1],
                "width": merged_array.shape[2],
                "transform": merged_transform,
            }
        )
    finally:
        # Close all datasets
        src = None
        for dataset in crs_datasets:
            dataset.close()

        if tmp_dir is not None:
            tmp_dir.cleanup()

    # Save merge datasets
    write(merged_array, merged_meta, merged_path, **write_kwargs)


def unpackbits(array: np.ndarray, nof_bits: int) -> np.ndarray:
    """
    Function found
    `here <https://stackoverflow.com/questions/18296035/how-to-extract-the-bits-of-larger-numeric-numpy-data-types>`_

    Args:
        array (np.ndarray): Array to unpack
        nof_bits (int): Number of bits to unpack

    Returns:
        np.ndarray: Unpacked array

    Example:
        >>> bit_array = np.random.randint(5, size=[3,3])
        array([[1, 1, 3],
               [4, 2, 0],
               [4, 3, 2]], dtype=uint8)
        >>>
        >>> # Unpack 8 bits (8*1, as itemsize of uint8 is 1)
        >>> unpackbits(bit_array, 8)
        array([[[1, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0, 0, 0]],
               [[0, 0, 1, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0]],
               [[0, 0, 1, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0]]], dtype=uint8)
    """
    xshape = list(array.shape)
    array = array.reshape([-1, 1])
    msk = 2 ** np.arange(nof_bits, dtype=array.dtype).reshape([1, nof_bits])
    uint8_packed = (array & msk).astype(bool).astype(np.uint8)
    try:
        unpacked = uint8_packed.reshape(xshape + [nof_bits])
    except IndexError:
        # Workaround for weird bug in reshape with dask
        unpacked = uint8_packed.compute().reshape(xshape + [nof_bits])

    return unpacked


def read_bit_array(
    bit_mask: np.ndarray, bit_id: Union[list, int]
) -> Union[np.ndarray, list]:
    """
    Read bit arrays as a succession of binary masks (sort of read a slice of the bit mask, slice number bit_id)

    Args:
        bit_mask (np.ndarray): Bit array to read
        bit_id (int): Bit ID of the slice to be read
          Example: read the bit 0 of the mask as a cloud mask (Theia)

    Returns:
        Union[np.ndarray, list]: Binary mask or list of binary masks if a list of bit_id is given

    Example:
        >>> bit_array = np.random.randint(5, size=[3,3])
        array([[1, 1, 3],
               [4, 2, 0],
               [4, 3, 2]], dtype=uint8)

        >>> # Get the 2nd bit array
        >>> read_bit_array(bit_array, 2)
        array([[0, 0, 0],
               [1, 0, 0],
               [1, 0, 0]], dtype=uint8)
    """
    # Get the number of bits
    nof_bits = 8 * bit_mask.dtype.itemsize

    # Read cloud mask as bits
    msk = unpackbits(bit_mask, nof_bits)

    # Only keep the bit number bit_id and reshape the vector
    if isinstance(bit_id, list):
        bit_arr = [msk[..., bid] for bid in bit_id]
    else:
        bit_arr = msk[..., bit_id]

    return bit_arr


@any_raster_to_rio_ds
def hillshade(
    ds: AnyRasterType, azimuth: float = 315, zenith: float = 45
) -> (np.ma.masked_array, dict):
    """
     Compute the hillshade of a DEM from an azimuth and zenith angle (in degrees).

     Goal: replace `gdaldem CLI <https://gdal.org/programs/gdaldem.html>`_

     NB: altitude = 90 - zenith

     .. WARNING::

         - It uses a 2nd order gradient instead of Horn's or Zevenbergen & Thorne's formula
         - z_factor is fixed to 1.0
         - scale managed by ds resolution

    `Reference <https://github.com/nasa/World-Wind-Java/blob/7c9886ab67ac03d53bdb04f161b9605d3f3dd810/GDAL/GDAL-1.7.2/apps/gdaldem.cpp#L349>`_

     Args:
         ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata
         azimuth (float): Azimuth of the light, in degrees. 0 if it comes from the top of the raster, 90 from the east, ...
         zenith (float): Zenith angle in degrees

     Returns:
         (np.ma.masked_array, dict): Hillshade and its metadata
    """
    array = ds.read(masked=True)

    # Squeeze if needed
    expand = False
    if len(array.shape) == 3 and array.shape[0] == 1:
        array = np.squeeze(array)  # Use this trick to make the sieve work
        expand = True

    # Compute angles
    az_rad = azimuth * DEG_2_RAD
    alt_rad = (90 - zenith) * DEG_2_RAD

    # Compute slope and aspect
    dx, dy = np.gradient(np.where(array.mask, 0.0, array.data), *ds.res)
    x2_y2 = dx**2 + dy**2
    aspect = np.arctan2(dx, dy)

    # Compute hillshade (GDAL algo)
    hshade = (
        np.sin(alt_rad) + np.cos(alt_rad) * np.sqrt(x2_y2) * np.sin(aspect - az_rad)
    ) / np.sqrt(1 + x2_y2)
    hshade = np.where(hshade <= 0, 1.0, 254.0 * hshade + 1)

    # Use this trick to get the array back to 'normal'
    if expand:
        hshade = np.expand_dims(hshade, axis=0)

    # Convert to masked array
    hillshade_msk = np.ma.masked_array(hshade, array.mask, fill_value=ds.nodata)

    # Meta
    meta = update_meta(hillshade_msk, ds.meta)

    return hillshade_msk, meta


@any_raster_to_rio_ds
def slope(
    ds: AnyRasterType,
    in_pct: bool = False,
    in_rad: bool = False,
) -> (np.ma.masked_array, dict):
    """
     Compute the slope of a DEM (in degrees).

     Goal: replace `gdaldem CLI <https://gdal.org/programs/gdaldem.html>`_

     .. WARNING::

         - It uses a 2nd order gradient instead of Horn's or Zevenbergen & Thorne's formula
         - z_factor is fixed to 1.0
         - scale managed by ds resolution

    `Reference <https://git.earthdata.nasa.gov/projects/GEE/repos/gdal-enhancements-for-esdis/browse/gdal-1.10.0/apps/gdaldem.cpp>`_

     Args:
         ds (AnyRasterType): Path to the raster, its dataset, its :code:`xarray` or a tuple containing its array and metadata
         in_pct (bool): Outputs slope in percents
         in_rad (bool): Outputs slope in radians. Not taken into account if :code:`in_pct == True`

     Returns:
         (np.ma.masked_array, dict): Slope and its metadata
    """
    array = ds.read(masked=True)

    # Squeeze if needed
    expand = False
    if len(array.shape) == 3 and array.shape[0] == 1:
        array = np.squeeze(array)  # Use this trick to make the sieve work
        expand = True

    # Compute slope (on unmasked data)
    dx, dy = np.gradient(np.where(array.mask, 0.0, array.data), *ds.res)
    x2_y2 = dx**2 + dy**2

    if in_pct:
        slp = 100 * (np.sqrt(x2_y2))
    else:
        slp = np.arctan(np.sqrt(x2_y2))

        # Convert into degrees
        if not in_rad:
            slp = slp / DEG_2_RAD

    # Use this trick to get the array back to 'normal'
    if expand:
        slp = np.expand_dims(slp, axis=0)

    # Convert to masked array
    slp_msk = np.ma.masked_array(slp, array.mask, fill_value=ds.nodata)

    # Meta
    meta = update_meta(slp_msk, ds.meta)

    return slp_msk, meta


def reproject_match(
    dst_meta: dict,
    src_arr: AnyNumpyArray,
    src_meta: dict,
    resampling: Resampling = Resampling.nearest,
    **kwargs,
) -> (AnyNumpyArray, dict):
    """
    Reproject a raster to match the resolution, projection, and region of another raster.

    Matching rioxarray reproject_match.

    Args:
        dst_meta (dict): Destination metadata
        src_arr (AnyNumpyArray): Source raster's array
        src_meta (dict): Source metadata
        resampling (Resampling): Resampling method
        **kwargs: Passing other kwargs to `calculate_default_transform` and `reproject`

    Returns:
        AnyNumpyArray, dict: Reprojected array and its metadata
    """

    # Source metadata
    src_crs = src_meta["crs"]
    src_count = src_meta["count"]
    src_dtype = src_meta["dtype"]
    src_nodata = src_meta["nodata"]

    # Destination metadata
    dst_crs = dst_meta["crs"]
    meta = dst_meta.copy()

    # Reproject data
    dst_arr = np.empty(
        (src_count, dst_meta["height"], dst_meta["width"]), dtype=src_dtype
    )
    dst_arr, dst_tr = warp.reproject(
        source=src_arr,
        destination=dst_arr,
        src_crs=src_crs,
        dst_crs=dst_crs,
        src_transform=src_meta["transform"],
        dst_transform=dst_meta["transform"],
        src_nodata=src_nodata,
        dst_nodata=dst_meta["nodata"],  # input data should be in integer
        num_threads=perf.get_max_cores(),
        resampling=resampling,
        **kwargs,
    )

    # Update metadata
    meta["count"] = src_count
    meta["nodata"] = src_nodata
    meta["dtype"] = src_dtype
    meta["transform"] = dst_tr
    meta["crs"] = dst_crs
    meta["driver"] = "GTiff"  # Force GTiff

    return dst_arr, meta



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing string functions"""

import argparse
import logging
from datetime import date, datetime

import pytest

from sertit import ci, strings

ci.reduce_verbosity()


def test_conversion():
    # Str to bool
    true_str = (True, "yes", "true", "t", "y", "1")
    false_str = (False, "no", "false", "f", "n", "0")
    for true, false in zip(true_str, false_str):
        assert strings.str_to_bool(true)
        assert not strings.str_to_bool(false)

    # Str to logging verbosity
    debug_str = ("debug", "d", 10)
    info_str = ("info", "i", 20)
    warn_str = ("warning", "w", "warn", 30)
    err_str = ("error", "e", "err", 40)
    for debug, info, warn, err in zip(debug_str, info_str, warn_str, err_str):
        assert strings.str_to_verbosity(debug) == logging.DEBUG
        assert strings.str_to_verbosity(info) == logging.INFO
        assert strings.str_to_verbosity(warn) == logging.WARNING
        assert strings.str_to_verbosity(err) == logging.ERROR

    with pytest.raises(argparse.ArgumentTypeError):
        strings.str_to_verbosity("oopsie")

    # Str to list of dates
    list_of_str_dates = "20200909105055, 2019-08-06;19560702121212\t2020-08-09"
    list_of_datetimes = [
        datetime(2020, 9, 9, 10, 50, 55),
        datetime(2019, 8, 6),
        datetime(1956, 7, 2, 12, 12, 12),
        datetime(2020, 8, 9),
    ]
    list_of_dates = [
        datetime(2020, 9, 9, 10, 50, 55),
        date(2019, 8, 6),
        datetime(1956, 7, 2, 12, 12, 12),
        date(2020, 8, 9),
    ]
    assert list_of_datetimes == strings.str_to_list_of_dates(
        list_of_str_dates, date_format="%Y%m%d%H%M%S", additional_separator="\t"
    )
    assert strings.str_to_list_of_dates(list_of_dates) == list_of_datetimes
    assert strings.str_to_list_of_dates(list_of_datetimes) == list_of_datetimes

    # Str to list
    list_str_up = ["A", "B"]
    list_str_low = ["a", "b"]
    assert strings.str_to_list(list_str_up) == list_str_up
    assert strings.str_to_list(list_str_low, case="upper") == list_str_up
    assert strings.str_to_list(list_str_up, case="lower") == list_str_low


def test_str():
    """Test string function"""
    tstr = "ThisIsATest"
    assert strings.snake_to_camel_case(strings.camel_to_snake_case(tstr)) == tstr
    assert strings.to_cmd_string(tstr) == f'"{tstr}"'


def test_uuid():
    """Test string function"""
    is_uuid_4 = "d0910ba4-9ed9-409a-8527-e40b581bf0c2"
    is_uuid_1 = "78d31fbe-2454-11ed-b256-7b4fd196c4c5"
    is_not_uuid = "random"

    assert strings.is_uuid(is_uuid_4, 4)

    for i in [1, 2, 3, 5]:
        assert not strings.is_uuid(is_uuid_4, i)

    assert not strings.is_uuid(is_uuid_1, 4)

    for i in [2, 3, 4, 5]:
        assert not strings.is_uuid(is_uuid_1, i)

    for i in range(1, 5):
        assert not strings.is_uuid(is_not_uuid, i)

    with pytest.raises(ValueError):
        assert not strings.is_uuid(is_uuid_4, 8)



from pathlib import Path
from typing import Union

import numpy as np
from cloudpathlib import CloudPath

from sertit import AnyPath
from sertit.types import AnyPathType, is_iterable, make_iterable


def test_types():
    """Test some type aliases"""
    assert AnyPathType == Union[Path, CloudPath]


def test_is_iterable():
    """Test is_iterable"""
    assert is_iterable((1, 2, 3))
    assert is_iterable([1, 2, 3])
    assert is_iterable({1, 2, 3})
    assert is_iterable(np.array([1, 2, 3]))
    assert not is_iterable("1, 2, 3")
    assert is_iterable("1, 2, 3", str_allowed=True)
    assert not is_iterable(1)
    assert not is_iterable(AnyPath("1, 2, 3"))


def test_make_iterable():
    """Test make_iterable"""

    def assert_mi(val, is_it=True, **kwargs):
        if not is_it:
            val = [val]

        comp = True if val is None else val == make_iterable(val, **kwargs)
        try:
            assert comp
        except ValueError:
            assert all(comp)

    assert_mi((1, 2, 3))
    assert_mi([1, 2, 3])
    assert_mi({1, 2, 3})
    assert_mi(np.array([1, 2, 3]))
    assert_mi("1, 2, 3", str_allowed=True)
    assert_mi("1, 2, 3", str_allowed=False, is_it=False)
    assert_mi(1, is_it=False)
    assert_mi(None, convert_none=True)
    assert_mi(None, convert_none=False, is_it=False)



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
S3 tools
"""

import logging
import os
from contextlib import contextmanager
from functools import wraps

from cloudpathlib import S3Client

from sertit.logs import SU_NAME

LOGGER = logging.getLogger(SU_NAME)

AWS_ACCESS_KEY_ID = "AWS_ACCESS_KEY_ID"
"""
Environment variable linked to AWS Access Key ID.
"""

AWS_SECRET_ACCESS_KEY = "AWS_SECRET_ACCESS_KEY"
"""
Environment variable linked to AWS Secret Access Key.
"""

AWS_S3_ENDPOINT = "AWS_S3_ENDPOINT"
"""
Environment variable linked to AWS endpoint.
"""

USE_S3_STORAGE = "USE_S3_STORAGE"
"""
Environment variable created to use Unistra's S3 bucket.
"""


def in_house_s3_configs():
    """
    This function declares in-house S3 configurations for rasterio, pyogrio and cloudpathlib.

    You can control how rasterio, pyogrio and cloudpathlib connects to S3 server by changing the input configurations bellow.
    Then, configuration files are translated for each library into their own configuration system.

    To get more information read the docstring of the function args_dict.

    Returns:

    """
    import rasterio

    args_rasterio_env = rasterio.env.getenv() if rasterio.env.hasenv() else {}
    args_rasterio_l = [
        {"kwargs": "profile_name"},
        {"name": "AWS_S3_ENDPOINT", "kwargs": "endpoint", "env": "AWS_S3_ENDPOINT"},
        {"name": "CPL_AWS_CREDENTIALS_FILE", "env": "AWS_SHARED_CREDENTIALS_FILE"},
        {"name": "AWS_CONFIG_FILE", "env": "AWS_CONFIG_FILE"},
        {
            "name": "CPL_CURL_VERBOSE",
            "default": args_rasterio_env.get("CPL_CURL_VERBOSE", False),
        },
        {
            "name": "GDAL_DISABLE_READDIR_ON_OPEN",
            "default": args_rasterio_env.get("GDAL_DISABLE_READDIR_ON_OPEN", False),
        },
        {"name": "AWS_NO_SIGN_REQUEST", "kwargs": "no_sign_request"},
        {"name": "AWS_REQUEST_PAYER", "kwargs": "requester_pays"},
    ]

    args_pyogrio_l = [
        {"name": "AWS_PROFILE", "kwargs": "profile_name", "env": "AWS_PROFILE"},
        {"name": "AWS_S3_ENDPOINT", "kwargs": "endpoint", "env": "AWS_S3_ENDPOINT"},
        {"name": "CPL_AWS_CREDENTIALS_FILE", "env": "AWS_SHARED_CREDENTIALS_FILE"},
        {"name": "AWS_CONFIG_FILE", "env": "AWS_CONFIG_FILE"},
        {"name": "AWS_NO_SIGN_REQUEST", "kwargs": "no_sign_request"},
        {"name": "AWS_REQUEST_PAYER", "kwargs": "requester_pays"},
    ]

    s3_client_args_l = [
        {
            "name": "endpoint_url",
            "kwargs": "endpoint",
            "env": "AWS_S3_ENDPOINT",
            "prefix": "https://",
        },
        {"kwargs": "aws_access_key_id", "env": "AWS_ACCESS_KEY_ID"},
        {"kwargs": "aws_secret_access_key", "env": "AWS_SECRET_ACCESS_KEY"},
        {"kwargs": "requester_pays"},
        {"kwargs": "no_sign_request", "default": False},
        {"kwargs": "profile_name", "env": "AWS_S3_PROFILE"},
        {"kwargs": "aws_session_token"},
        {"kwargs": "botocore_session"},
        {"kwargs": "boto3_session"},
        {"kwargs": "file_cache_mode"},
        {"kwargs": "local_cache_dir"},
        {"kwargs": "boto3_transfer_config"},
        {"kwargs": "content_type_method"},
        {"kwargs": "extra_args"},
    ]
    return {
        "cloudpathlib": s3_client_args_l,
        "pyogrio": args_pyogrio_l,
        "rasterio": args_rasterio_l,
    }


def s3_env(*args, **kwargs):
    """
    Create S3 compatible storage environment
    You need to set endpoint url if you use s3 compatible storage
    since GDAL/Rasterio does not read endpoint url from config file.

    This function searches for S3 configuration in many places.
    It does apply configuration variables precedence, and you might have a use for it.
    Here is the order of precedence from least to greatest
    (the last listed configuration variables override all other variables):

    1. AWS profile
    2. Given arguments
    3. AWS environment variable

    Returns:
        Callable: decorated function

    Example:
        >>> from sertit.s3 import s3_env
        >>> from sertit import AnyPath
        >>> @s3_env(endpoint="s3.unistra.fr")
        >>> def file_exists(path: str):
        >>>     pth = AnyPath(path)
        >>>     print(pth.exists())
        >>> file_exists("s3://sertit-geodatastore/GLOBAL/COPDEM_30m/COPDEM_30m.vrt")
        True
    """
    import rasterio
    from pyogrio import set_gdal_config_options

    use_s3 = kwargs.get("use_s3_env_var", USE_S3_STORAGE)
    args = s3_args(*args, **kwargs)
    args_rasterio = args["rasterio"]
    args_pyogrio = args["pyogrio"]

    def decorator(function):
        @wraps(function)
        def s3_env_wrapper(*_args, **_kwargs):
            """S3 environment wrapper"""
            if int(os.getenv(use_s3, 1)):
                # Define S3 client for S3 paths
                define_s3_client(**kwargs)
                os.environ[use_s3] = "1"
                LOGGER.info("Using S3 files")
                set_gdal_config_options(args_pyogrio)

                with rasterio.Env(**args_rasterio):
                    return function(*_args, **_kwargs)

            else:
                os.environ[use_s3] = "0"
                LOGGER.info("Using on disk files")
                return function(*_args, **_kwargs)

        return s3_env_wrapper

    return decorator


@contextmanager
def temp_s3(
    endpoint: str = None,
    profile_name: str = None,
    requester_pays: bool = False,
    no_sign_request: bool = False,
    **kwargs,
) -> None:
    """
    Initialize a temporary S3 environment as a context manager
    You need to set endpoint url if you use s3 compatible storage
    since GDAL/Rasterio does not read endpoint url from config file.

    This function searches for S3 configuration in many places.
    It does apply configuration variables precedence, and you might have a use for it.
    Here is the order of precedence from least to greatest
    (the last listed configuration variables override all other variables):

    1. AWS profile
    2. Given arguments
    3. AWS environment variable

    Args:
        endpoint: Endpoint to s3 path in the form s3.yourdomain.com
        profile_name: The name of your AWS profile
        requester_pays (bool): True if the endpoint says 'requester pays'
        no_sign_request (bool): True if the endpoint is open access

    Example:
        >>> from sertit.s3 import temp_s3
        >>> from sertit import AnyPath
        >>> def file_exists(path: str):
        >>>     with temp_s3(endpoint="s3.unistra.fr"):
        >>>         pth = AnyPath(path)
        >>>         print(pth.exists())
        >>> file_exists("s3://sertit-geodatastore/GLOBAL/COPDEM_30m/COPDEM_30m.vrt")
        True
    """
    import rasterio
    from pyogrio import set_gdal_config_options

    kwargs_cp = kwargs.copy()
    kwargs_cp["endpoint"] = endpoint
    kwargs_cp["profile_name"] = profile_name
    kwargs_cp["requester_pays"] = requester_pays
    kwargs_cp["no_sign_request"] = no_sign_request
    args = s3_args(**kwargs_cp)
    args_rasterio = args["rasterio"]
    args_pyogrio = args["pyogrio"]

    # Define S3 client for S3 paths
    try:
        set_gdal_config_options(args_pyogrio)
        with rasterio.Env(**args_rasterio):
            yield define_s3_client(
                endpoint=endpoint,
                profile_name=profile_name,
                requester_pays=requester_pays,
                no_sign_request=no_sign_request,
                **kwargs,
            )
    finally:
        # Clean env
        S3Client().set_as_default_client()


def define_s3_client(
    endpoint=None,
    profile_name=None,
    requester_pays: bool = False,
    no_sign_request: bool = False,
    **kwargs,
):
    """
    Define S3 client
    This function searches for S3 configuration in many places.
    It does apply configuration variables precedence, and you might have a use for it.
    Here is the order of precedence from least to greatest
    (the last listed configuration variables override all other variables):

    1. AWS profile
    2. Given arguments
    3. AWS environment variable

    Args:
        endpoint: The s3 endpoint (s3.yourdomain.com)
        profile_name: The name of the aws profile. Default to default profile in AWS configuration file.
        requester_pays (bool): True if the endpoint says 'requester pays'
        no_sign_request (bool): True if the endpoint is open access
    """
    kwargs_cp = kwargs.copy()
    kwargs_cp["endpoint"] = endpoint
    kwargs_cp["profile_name"] = profile_name
    kwargs_cp["requester_pays"] = requester_pays
    kwargs_cp["no_sign_request"] = no_sign_request

    args = s3_args(**kwargs_cp)
    args_s3_client = args["cloudpathlib"]
    client = S3Client(**args_s3_client)

    client.set_as_default_client()


def s3_args(*args, **kwargs) -> dict:
    """
    This function returns ready to use configurations for rasterio, pyogrio and cloudpathlib.
    For each in-house input configurations, it applies the function args_dict.

    Args:
        *args:
        **kwargs:

    Returns:

    """
    import rasterio

    in_house_configs = in_house_s3_configs()

    s3_client_args_l = in_house_configs["cloudpathlib"]
    args_pyogrio_l = in_house_configs["pyogrio"]
    args_rasterio_l = in_house_configs["rasterio"]

    s3_client_args = args_dict(s3_client_args_l, kwargs)
    args_pyogrio = args_dict(args_pyogrio_l, kwargs)

    args_rasterio = args_dict(args_rasterio_l, kwargs)
    args_rasterio_env = rasterio.env.getenv() if rasterio.env.hasenv() else {}
    args_rasterio_env.update(args_rasterio)

    return {
        "cloudpathlib": s3_client_args,
        "pyogrio": args_pyogrio,
        "rasterio": args_rasterio,
    }


def args_dict(args_l: list[dict], kwargs) -> dict:
    """
    This function converts a single in-house S3 configuration to a dictionary containing ready to use key/value parameters.

    The input is a list of dict to process. Each dictionary can contain the following key:
    - name: The name of the key in the output dict. If not given, the name of the key is the value of kwargs.
    - kwargs: The name of the kwargs argument whose value is taken to set the value in output dict.
    - env: The name of the envrionment variable whose value is taken to set the value in output dict.
    - prefix: A prefix prefixed to the value in the ouput dict.

    For example, if the environment variable "AWS_PROFILE" is unset, the following input:
    args_l = [{"name": "AWS_PROFILE", "kwargs": "profile_name", "env": "AWS_PROFILE"}], kwargs={"profile_name": "unistra"}
    will output:
    {"AWS_PROFILE": "unistra"}

    Here is the order of precedence from least to greatest
    (the last listed configuration variables override all other variables):

    1. Value from kwargs.
    2. Default value from key "default".
    3. Value from environment variable.

    If no value is found, the output dict will not contain the wanted parameter.

    Args:
        args_l: A list of parameters to extract. Each element will give a single key/value parameter.
        kwargs: Considered kwargs input to set parameters.

    Returns:

    """
    ret = {}
    for arg in args_l:
        arg_name = arg["name"] if "name" in arg else arg["kwargs"]

        # First, set with default value
        arg_value = arg.get("default")
        if arg_value is not None:
            ret[arg_name] = arg_value
        # Override with kwargs
        arg_value = kwargs.get(arg["kwargs"]) if arg.get("kwargs") is not None else None
        if arg_value is not None:
            ret[arg_name] = arg_value
        # Override with environment variable
        arg_value = os.getenv(arg.get("env")) if arg.get("env") is not None else None
        if arg_value is not None and arg_value != "":
            ret[arg_name] = arg_value

        if arg.get("prefix") is not None and ret.get(arg_name) is not None:
            ret[arg_name] = arg.get("prefix") + ret[arg_name]

    # Some exceptions
    if ret.get("AWS_NO_SIGN_REQUEST") is not None:
        ret["AWS_NO_SIGN_REQUEST"] = "YES" if ret["AWS_NO_SIGN_REQUEST"] else "NO"
    if ret.get("AWS_REQUEST_PAYER") is not None:
        ret["AWS_REQUEST_PAYER"] = "requester" if ret["AWS_REQUEST_PAYER"] else None

    if ret.get("requester_pays") is not None:
        if ret.get("extra_args") is not None:
            ret["extra_args"].update({"RequestPayer": "requester"})
        else:
            ret["extra_args"] = {"RequestPayer": "requester"}
        ret.pop("requester_pays")
    return ret



import contextlib
import logging
import os
from contextlib import contextmanager

import psutil
import xarray as xr

from sertit import logs
from sertit.types import AnyXrDataStructure

LOGGER = logging.getLogger(logs.SU_NAME)

DEFAULT_CHUNKS = "auto"
""" Default chunks used in Sertit library (if dask is installed) """

SERTIT_DEFAULT_CHUNKS = "SERTIT_DEFAULT_CHUNKS"
"""
Environment variable to override default chunks.

Available keywords (case agnostic): 
- Give :code:`NONE` to set :code:`None`
- Give :code:`TRUE` to set :code:`True`
- Give :code:`AUTO` to set :code:`"auto"`
"""


def is_dask_installed():
    try:
        from dask import optimize  # noqa: F401
        from dask.distributed import get_client  # noqa: F401

        return True
    except ModuleNotFoundError:
        return False


def get_client():
    client = None
    if is_dask_installed():
        from dask.distributed import get_client

        with contextlib.suppress(ValueError):
            # Return default client
            client = get_client()
    else:
        LOGGER.warning(
            "Can't import 'dask'. If you experiment out of memory issue, consider installing 'dask'."
        )

    return client


@contextmanager
def get_or_create_dask_client(processes=False, env_vars=None):
    """
    Return default Dask client or create a local cluster and linked  client if not existing
    Returns:
    """
    client = None
    try:
        if is_dask_installed():
            from dask.distributed import Client, get_client

            try:
                # Return default client
                client = get_client()
            except ValueError:
                if processes:
                    # Gather information to create a client adapted to the computer
                    ram_info = psutil.virtual_memory()
                    available_ram = ram_info.available / 1024 / 1024 / 1024
                    available_ram = 0.9 * available_ram

                    n_workers = 1
                    memory_limit = f"{available_ram}Gb"
                    if available_ram >= 16:
                        n_workers = available_ram // 16
                        memory_limit = f"{16}Gb"

                    # Create a local cluster and return client
                    LOGGER.warning(
                        f"Init local cluster with {n_workers} workers and {memory_limit} per worker"
                    )
                    client = Client(
                        n_workers=int(n_workers),
                        threads_per_worker=4,
                        memory_limit=memory_limit,
                    )
                else:
                    # Create a local cluster (threaded)
                    LOGGER.warning("Init local cluster (threaded)")
                    client = Client(
                        processes=processes,
                    )
            if env_vars:
                client.run(
                    lambda: os.environ.update({k: v for k, v in env_vars.items()})
                )

            yield client

        else:
            LOGGER.warning(
                "Can't import 'dask'. If you experiment out of memory issue, consider installing 'dask'."
            )
    finally:
        try:
            if client is not None:
                client.close()
        except Exception as ex:
            LOGGER.warning(ex)


def get_dask_lock(name):
    """
    Get a dask lock with given name.
    This lock is used in rioxarray.to_raster (see https://corteva.github.io/rioxarray/stable/examples/dask_read_write.html)

    If a Multiple worker client exists: returns a distributed.Lock()
    Elif a Multithreaded client exists: returns a threading.Lock()
    else: returns None


    Args:
        name: The name of the lock
    Returns:
    """
    lock = None
    if is_dask_installed():
        from dask.distributed import Lock

        current_client = get_client()
        if current_client:
            lock = Lock(name, client=current_client)
    else:
        LOGGER.warning(
            "Can't import 'dask'. If you experiment out of memory issue, consider installing 'dask'."
        )
    return lock


def is_chunked(array: AnyXrDataStructure) -> bool:
    """
    Returns true if the array is still chunked.
    (i.e. its data is not computed, bnot loaded into memory as a numpy array)

    Args:
        array (AnyXrDataStructure): Array to check

    Returns: True if array is still chunked
    """
    try:
        if isinstance(array, xr.DataArray):
            is_chunked = array.chunks is not None
        else:
            is_chunked = len(array.chunks) > 0
    except AttributeError:
        is_chunked = False
    return is_chunked


def is_computed(array: AnyXrDataStructure) -> bool:
    """
    Returns true if the array is still chunked.
    (i.e. its data is not computed, bnot loaded into memory as a numpy array)

    Args:
        array (AnyXrDataStructure): Array to check

    Returns: True if array is still chunked
    """
    return not is_chunked(array)


def get_default_chunks():
    """
    Get the default chunks:

    - check if dask is available
    - check :code:`SERTIT_DEFAULT_CHUNKS` env variable
    - defaults on DEFAULT_CHUNKS
    """
    chunks = None
    if is_dask_installed():
        chunks = os.getenv(SERTIT_DEFAULT_CHUNKS, DEFAULT_CHUNKS)
        if chunks.lower() == "none":
            chunks = None
        elif chunks.lower() == "auto":
            chunks = "auto"
        elif chunks.lower() == "true":
            chunks = True

    return chunks


# From xarray: https://github.com/pydata/xarray/blob/30743945538ca2d276fc28eb221afa7bcb03978a/xarray/tests/__init__.py#L236-L259
class _CountingScheduler:
    """Simple dask scheduler counting the number of computes.

    Reference: https://stackoverflow.com/questions/53289286/"""

    def __init__(
        self,
        max_computes=0,
        nof_computes=None,
        dont_raise=False,
        force_synchronous=False,
    ):
        self.total_computes = 0
        self.nof_computes = nof_computes

        if max_computes is None or (
            nof_computes is not None and max_computes < nof_computes
        ):
            max_computes = nof_computes

        self.max_computes = max_computes
        self.dont_raise = dont_raise
        self.debug = force_synchronous

        # In case of debug, use the dask basic scheduler
        if self.debug:
            import dask

            self.get = dask.get
        else:
            self.get = None

    def __call__(self, dsk, keys, **kwargs):
        self.total_computes += 1

        # Log where the compute happens
        import traceback

        for tb in traceback.extract_stack()[::-1]:
            # Change this condition if we are using dask elsewhere than rasters.py and we want to display the tb
            if (
                tb.filename.lower().startswith("/home/data")
                and "/rasters.py" in tb.filename
            ):
                LOGGER.debug(
                    f"Computation number {self.total_computes}: {tb.line} | {tb.name} in {tb.filename} at line {tb.lineno}"
                )
                break

        # Raise or warn if too many computes have occurred
        if self.total_computes > self.max_computes:
            text = f"Too many computes. Total: {self.total_computes} > max: {self.max_computes}."
            if self.dont_raise:
                LOGGER.warning(text)
            else:
                raise RuntimeError(text)

        # Use the wanted get
        if self.get is None:
            client = get_client()
            if client:
                self.get = client.get

        if self.get is not None:
            return self.get(dsk, keys, **kwargs)

    def check_total_nof_computes(self):
        if self.nof_computes is not None and self.total_computes != self.nof_computes:
            text = f"Unexpected number of computes. Total: {self.total_computes} != {self.nof_computes}."

            if self.dont_raise:
                LOGGER.warning(text)
            else:
                raise RuntimeError(text)
        return True


@contextmanager
def raise_if_dask_computes(
    max_computes=0, nof_computes=None, dont_raise=False, force_synchronous=False
):
    # return a dummy context manager so that this can be used for non-dask objects
    if not is_dask_installed():
        yield contextlib.nullcontext()

    import dask

    scheduler = _CountingScheduler(
        max_computes=max_computes,
        nof_computes=nof_computes,
        dont_raise=dont_raise,
        force_synchronous=force_synchronous,
    )
    try:
        yield dask.config.set(scheduler=scheduler)
    finally:
        scheduler.check_total_nof_computes()
        # Make sure the counting scheduler is removed
        dask.config.set(scheduler=None)



import logging
import logging.handlers

from sertit.logs import deprecation_warning

# Arcpy types from inside a schema
SHORT = "int32:4"
""" 'Short' type for ArcGis GDB """

LONG = "int32:10"
""" 'Long' type for ArcGis GDB """

FLOAT = "float"
""" 'Float' type for ArcGis GDB """

DOUBLE = "float"
""" 'Double' type for ArcGis GDB """

TEXT = "str:255"
""" "Text" type for ArcGis GDB """

DATE = "datetime"
""" 'Date' type for ArcGis GDB """


# flake8: noqa
def init_conda_arcpy_env():
    """
    Initialize conda environment with Arcgis Pro.

    Resolves several issues.
    """
    try:
        from packaging.version import InvalidVersion, Version

        try:
            import fiona
            from fiona import Env as fiona_env

            with fiona_env():
                gdal_version = fiona.env.get_gdal_release_name()
                Version(gdal_version)
        except InvalidVersion:
            # workaround to https://community.esri.com/t5/arcgis-pro-questions/arcgispro-py39-gdal-version-3-7-0e-is-recognized/m-p/1364021
            import geopandas as gpd

            gpd.options.io_engine = "pyogrio"
    except ModuleNotFoundError:
        pass


class ArcPyLogger:
    def __init__(self, name=None, prefix_log_file="atools_"):
        """
        This class inits a ready to use python logger for ArcGis pro tool. Be sure that arcpy has been imported
        before using this class. It uses logging under the hood.
        It writes outputs to a temporary file and to the ArcGis console.
        The temporary file is removed when the user closes ArcGis.

        If you need a logger in an outside module or function, use :code:`logging.getLogger(LOGGER_NAME)`
        to get your logger.

        Args:
            name (str) : The name of the logger
            prefix_log_file (str) : The log filename is random, but you can prefix a name.
                The default value is "{{ name }}_".

        Example:
            >>> import logging
            >>> from sertit.arcpy import ArcPyLogger
            >>> arcpy_logger = ArcPyLogger(name="MyArcgisTool")
            Outputs written to file: C:\\Users\\bcoriat\\AppData\\Local\\Temp\\ArcGISProTemp15788\\MyArcgisTool_1bv0c1cl
            >>> logger = logging.getLogger("MyArcgisTool")
            >>> logger.info("Hello World !")
            Hello World !

        Warning:
            Python must keep a reference to the instantiated object during the execution of your program.
            That's why you must init this class once at the top level of your project.

            This will not work because Python destroys the object class.

            >>> ArcPyLogger(name="MyArcgisTool")
            >>> logger = logging.getLogger("MyArcgisTool")
            >>> logger.info("Hello World !")
        """
        self.name = name
        self.logger = None
        self.handler = None
        if name:
            self.prefix = name + "_"
        else:
            self.prefix = prefix_log_file
        self._set_logger()

    def __del__(self):
        self.logger.removeHandler(self.handler)

    def _set_logger(self):
        import tempfile

        logger = logging.getLogger(self.name)
        f = tempfile.NamedTemporaryFile(prefix=self.prefix, delete=False)

        # Create handler
        max_file_size = 1024 * 1024 * 2  # 2MB log files
        self.handler = ArcPyLogHandler(
            f.name,
            maxBytes=max_file_size,
            backupCount=10,
            encoding="utf-8",
        )
        logger.addHandler(self.handler)

        # Set formatter to handler
        formatter = logging.Formatter("%(levelname)-8s %(message)s")
        self.handler.setFormatter(formatter)

        # Set logger
        logger.setLevel(logging.DEBUG)
        self.logger = logger
        self.logger.info("You can read logs in the file: " + f.name)


class ArcPyLogHandler(logging.handlers.RotatingFileHandler):
    """
    Custom logging class that bounces messages to the arcpy tool window as well
    as reflecting back to the file.
    """

    def emit(self, record):
        """
        Write the log message
        """

        import arcpy

        try:
            msg = record.msg % record.args
        except:
            try:
                msg = record.msg.format(record.args)
            except:
                msg = record.msg

        if record.levelno >= logging.ERROR:
            arcpy.AddError(msg)
        elif record.levelno >= logging.WARNING:
            arcpy.AddWarning(msg)
        elif record.levelno >= logging.INFO:
            arcpy.AddMessage(msg)

        super(ArcPyLogHandler, self).emit(record)


def feature_layer_to_path(feature_layer) -> str:
    """
    .. deprecated:: 1.36.0
       Use :py:func:`gp_layer_to_path` instead.

    Use :func:`gp_layer_to_path` instead.

    Convert a feature layer to its source path.

    Args:
        feature_layer: Feature layer

    Returns:
        str: Path to the feature layer source

    """
    deprecation_warning("This function is deprecated. Use gp_layer_to_path instead.")
    # Get path
    if hasattr(feature_layer, "dataSource"):
        path = feature_layer.dataSource
    else:
        path = str(feature_layer)

    return path


def gp_layer_to_path(feature_layer) -> str:
    """
    Convert a GP layer to its source path.

    A GP layer in ArcGis is a layer in the content panel. Thus, the user can simply choose the layer in a dropdown menu.
    This function adds the possibility to get the source path of this layer if the user chose in the dropdown menu or
    drag and drop from the Windows explorer.

    Args:
        feature_layer: Feature layer or Raster layer

    Returns:
        str: Path to the feature or raster layer source

    Examples:
        For python toolbox, in the getParameterInfo() method use GPLayer, GPFeatureLayer or GPRasterLayer datatype.

        For vector layer use GPFeatureLayer:

        >>> import arcpy
        >>> aoi = arcpy.Parameter(
        >>>    displayName="Aoi",
        >>>    name="aoi",
        >>>    datatype="GPFeatureLayer",
        >>>    parameterType="Required",
        >>>    direction="Input",
        >>> )

        For raster layer, use GPRasterLayer:

        >>> import arcpy
        >>> nir_path = arcpy.Parameter(
        >>>    displayName="Nir infrared band",
        >>>    name="nir_path",
        >>>    datatype="GPRasterLayer",
        >>>    parameterType="Optional",
        >>>    direction="Input",
        >>> )

        If your layer may be a feature or raster layer, use GPLayer:

        >>> import arcpy
        >>> dem_path = arcpy.Parameter(
        >>>    displayName="DEM path as isoline or raster",
        >>>    name="dem_path",
        >>>    datatype="GPLayer",
        >>>    parameterType="Optional",
        >>>    direction="Input",
        >>> )

        Then in the execute() method, you can use this function to retrieve the real path to the layer.

        >>> aoi_path = feature_layer_to_path(parameters[0].value)
        >>> print(aoi_path)
        D:/data/project/aoi/aoi.shp

    """
    # Get path
    if hasattr(feature_layer, "dataSource"):
        path = feature_layer.dataSource
    else:
        path = str(feature_layer)

    return path
